import isopy

# This will be deprecated in favour of Ruthenium
# But can stay until that package is ready
def check_type(name, value, *accepted_types, coerce=False, coerce_into=None, allow_none=False):
    if allow_none and value is None:
        return value

    for accepted_type in accepted_types:
        if isinstance(value, accepted_type):
            return value
    if coerce:
        if coerce_into is None:
            coerce_into = accepted_types
        elif not isinstance(coerce_into, list):
            coerce_into = [coerce_into]
        for coerce_type in coerce_into:
            try:
                return coerce_type(value)
            except:
                continue
        raise TypeError('parameter "{}": unable to coerce \'{}\' into \'{}\''.format(name,
                                                                                     value.__class__.__name__, '\' or \''.join([ct.__name__ for ct in coerce_into])))
    else:
        raise TypeError('parameter "{}": must be \'{}\' not \'{}\''.format(name,
                            '\' or \''.join([ct.__name__ for ct in coerce_into], value.__class__.__name__)))


import isopy
from isopy import core
from isopy import io
import numpy as np
import functools
import os

########################
### Reference Values ###
########################

def load_refval_values(filename, keys_in_first ='r'):
    filepath = os.path.join(os.path.dirname(__file__), 'referencedata', f'{filename}.csv')
    data = io.read_csv(filepath, keys_in_first=keys_in_first)
    data = {key: value[0] for key, value in data.items()}

    return data

# Assumes all default and reference values are properties
def is_default_value(func):
    setattr(func.fget, '_defval', True)
    setattr(func.fget, '_descr', func.fget.__doc__.lstrip('\n').split('\n', 1)[0].strip())
    return func


class RefValGroup:
    def __init__(self, parent):
        self._parent = parent

    def __repr__(self):
        names = "', '".join(self.ls())
        return f"isopy.refval.{self.__class__.__name__}(['{names}'])"

    def _repr_markdown_(self):
        descriptions = []
        for name in self.ls():
            doc = self.__class__.__dict__[name].__doc__
            doc = doc.split('\n\n', 1)[0].replace('\n', '').strip()
            descriptions.append(f'* **{name}** - {doc}')

        descriptions = '\n'.join(descriptions)
        return f'**isopy.refval.{self.__class__.__name__}** contains the following reference values\n{descriptions}'

    def ls(self):
        """
        Returns a list will all reference values in this group.
        """
        defval = []
        refval = []
        for name, item in self.__class__.__dict__.items():
            if type(item) is property:
                if getattr(item.fget, '_defval', False):
                    defval.append(name)
                else:
                    refval.append(name)
            
        return defval + refval

class mass(RefValGroup):
    @core.cached_property
    def isotopes(self):
        """
        Dictionary containing all naturally occuring isotopes with a given mass number.

        The dictionary is constructed from the isotopes in
        :attr:`isotope.best_abundance_measurement_M16`.

        The ``get()`` method of this dictionary will
        return an empty :class:`IsotopeKeyList` for absent keys.

        Examples
        --------
        >>> isopy.refval.mass.isotopes[40]
        IsotopeKeyList('40Ar', '40K', '40Ca')

        >>> isopy.refval.mass.isotopes.get(96)
        IsotopeKeyList('96Zr', '96Mo', '96Ru')
        """

        fraction = isopy.refval.isotope.best_measurement_fraction_M16
        isotopes = {f'{mass}':
                    isopy.askeylist([k for k in fraction if k._filter_(mass_number = {'key_eq': (mass,)})])
                    for mass in range(1, 239)}
        return core.IsopyDict(**{key: value for key, value in isotopes.items() if len(value) > 0},
                               default_value=isopy.askeylist(), readonly=True)

class element(RefValGroup):
    @core.cached_property
    def isotopes(self):
        """
        Dictionary containing all naturally occurring isotopes for each element symbol.

        The dictionary is constructed from the isotopes in
        :attr:`isotope.best_abundance_measurement_M16`.

        The ``get()`` method of this dictionary will
        return an empty :class:`IsotopeKeyList` for absent keys.

        Examples
        --------
        >>> isopy.refval.element.isotopes['pd']
        IsotopeKeyList('102Pd', '104Pd', '105Pd', '106Pd', '108Pd', '110Pd')

        >>> isopy.refval.element.isotopes.get('ge')
        IsotopeKeyList('70Ge', '72Ge', '73Ge', '74Ge', '76Ge')
        """

        fraction = isopy.refval.isotope.best_measurement_fraction_M16
        return core.IsopyDict(**{symbol: isopy.askeylist([k for k in fraction if k._filter_(element_symbol = {'key_eq': (symbol,)})])
                               for symbol in self.all_symbols}, default_value=isopy.askeylist(), readonly=True)

    @core.cached_property
    def all_symbols(self):
        """
        A tuple of all the element symbols.

        Examples
        --------
        >>> isopy.refval.element.all_symbols[:5] # first 5 element symbols
        (ElementKeyString('H'), ElementKeyString('He'), ElementKeyString('Li'),
         ElementKeyString('Be'), ElementKeyString('B'))
        """
        return tuple(self.symbol_name.keys())

    @core.cached_property
    def symbol_name(self):
        """
        Dictionary containing the full element name mapped to the element symbol.

        The first letter of the element name is capitalised.

        The ``get()`` method of this dictionary will
        return ``None`` for absent keys.

        Examples
        --------
        >>> isopy.refval.element.symbol_name['pd']
        'Palladium'

        >>> isopy.refval.element.symbol_name.get('ge')
        'Germanium'
        """
        return core.IsopyDict(load_refval_values('element_symbol_name').items(), readonly = True)

    @core.cached_property
    def name_symbol(self):
        """
        Dictionary containing the element symbol mapped to the full element name.

        The first letter of the element name must be capitalised.

        The ``get()`` method of this dictionary will
        return ``None`` for absent keys.

        Examples
        --------
        >>> isopy.refval.element.symbol_name['Palladium']
        ElementKeyString('Pd')

        >>> isopy.refval.element.symbol_name.get('Germanium')
        ElementKeyString('Ge')
        """
        return dict({name: symbol for symbol, name in self.symbol_name.items()}, readonly = True)

    @core.cached_property
    def atomic_weight(self):
        """
        Dictionary containing the atomic weight for each element symbol.

        The atomic weights are calculated using the isotopic abundances from
        :attr:`isotope.best_abundance_measurement_M16` and the isotopic masses from
        :attr:`isotope.isotope.mass_W17`.

        The ``get()`` method of this dictionary will
        return ``None`` for absent keys.

        Examples
        --------
        >>> isopy.refval.element.atomic_weight['pd']
        106.41532788648

        >>> isopy.refval.element.atomic_weight.get('ge')
        72.6295890304831
        """
        weights = {}
        # Isotopes is also taken from best_measurement_fraction_M16
        for element, isotopes in self.isotopes.items():
            weights[element] = np.sum([self._parent.isotope.mass_W17.get(isotope) *
                                       self._parent.isotope.best_measurement_fraction_M16.get(isotope)
                                       for isotope in isotopes])
        return core.RefValDict(**weights, readonly=True, ratio_function=np.divide,
                               molecule_functions='abundance')

    @core.cached_property
    def atomic_number(self):
        """
        Dictionary containing the atomic number for each element symbol.

        The ``get()`` method of this dictionary will
        return ``None`` for absent keys.

        Examples
        --------
        >>> isopy.refval.element.atomic_number['pd']
        46

        >>> isopy.refval.element.atomic_number.get('ge')
        32
        """
        return core.RefValDict(load_refval_values('element_atomic_number'), readonly = True)

    @core.cached_property
    def initial_solar_system_abundance_L09(self):
        """
        Dictionary containing the element abundance of the initial solar system composition (normalized to N(Si) = 10^6 atoms) from Lodders et al. (2019).

        Reference: `Lodders et al. (2019) <http://materials.springer.com/lb/docs/sm_lbs_978-3-540-88055-4_34>`_.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.
        """
        return core.RefValDict(load_refval_values('element_initial_solar_system_abundance_L09'),
                               default_value=np.nan, readonly = True, ratio_function=np.divide,
                               molecule_functions='abundance')

class isotope(RefValGroup):
    def __init__(self, parent):
        self._parent = parent
        self.__mass = None
        self.__fraction = None
    
    @is_default_value
    @property
    def mass(self):
        """
        Dictionary containing the default mass of each isotope.

        This is the default dictionary used in functions where the isotope mass is required.

        This dictionary is initially a copy of :attr:`isotope.mass_W17` but can be changed by the
        user. **Note** the default values used can change in future versions of isopy as newer
        values become avaliable.

        Examples
        --------
        >>> isopy.refval.isotope.mass['pd105']
        104.9050795

        >>> isopy.refval.isotope.mass.get('ge76')
        75.92140273

        >>> isopy.refval.isotope.mass.get('pd108/pd105')
        1.0285859589859039
        """
        if not self.__mass: self.__mass = self.mass_AME20
        return self.__mass

    @mass.setter
    def mass(self, value):
        if not isinstance(value, core.RefValDict):
            raise TypeError('attribute must be a scalar dictionary')
        self.__mass = value
    
    @is_default_value
    @property
    def fraction(self):
        """
        Dictionary containing the default fraction of each isotope.

        This is the default dictionary used in functions where the isotope abundance is required.

        This dictionary is initially a copy of :attr:`isotope.best_measurement_fracton_M16`
        but can be changed by the user. **Note** the default values used can change in future
        versions of isopy as newer values become avaliable.

        Examples
        --------
        >>> isopy.refval.isotope.best_measurement_fraction_M16['pd105']
        0.2233

        >>> isopy.refval.isotope.best_measurement_fraction_M16.get('ge76')
        0.07745

        >>> isopy.refval.isotope.best_measurement_fraction_M16.get('pd108/pd105')
        1.1849529780564263
        """
        if not self.__fraction: self.__fraction = self.best_measurement_fraction_M16
        return self.__fraction

    @fraction.setter
    def fraction(self, value):
        if not isinstance(value, core.RefValDict):
            raise TypeError('attribute must be a dictionary')
        self.__fraction = value

    @core.cached_property
    def mass_W17(self):
        """
        Dictionary containing isotope mass of each isotope from Wang et al. (2016).

        Reference: `Wang et al. (2016) <https://doi.org/10.1088/1674-1137/41/3/030003>`_.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.

        Examples
        --------
        >>> isopy.refval.isotope.mass_W17['pd105']
        104.9050795

        >>> isopy.refval.isotope.mass_W17.get('ge76')
        75.92140273

        >>> isopy.refval.isotope.mass_W17.get('pd108/pd105')
        1.0285859589859039
        """
        return core.RefValDict(load_refval_values('isotope_mass_W17'),
                               default_value=np.nan, readonly = True, ratio_function=np.divide,
                               molecule_functions='mass')

    @core.cached_property
    def mass_AME20(self):
        """
        Dictionary containing isotope mass of each isotope from the 2020 Atomic Mass Evaluation.

        Reference: `Atomic Mass Evaluation 2020 <https://www-nds.iaea.org/amdc/>`_.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.

        """
        return core.RefValDict(load_refval_values('isotope_mass_AME20', keys_in_first='c'),
                               default_value=np.nan, readonly=True, ratio_function=np.divide,
                               molecule_functions='mass')

    @core.cached_property
    def mass_number(self):
        """
        Dictionary containing mass number of each isotope.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.

        Examples
        --------
        >>> isopy.refval.isotope.mass_number['pd105']
        105

        >>> isopy.refval.isotope.mass_number.get('ge76')
        76

        >>> isopy.refval.isotope.mass_number.get('pd108/pd105')
        1.0285714285714285
        """
        return core.RefValDict({key: int(key.mass_number) for key in self.mass_W17},
                               default_value=np.nan, readonly=True, ratio_function=np.divide,
                               molecule_functions='mass')

    @core.cached_property
    def best_measurement_fraction_M16(self):
        """
        Dictionary containing the isotope fraction from the best avaliable measurement from Meija et al. (2016).

        Reference: `Meija et al. (2016) <https://doi.org/10.1515/pac-2015-0503>`_.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.

        Examples
        --------
        >>> isopy.refval.isotope.best_measurement_fraction_M16['pd105']
        0.2233

        >>> isopy.refval.isotope.best_measurement_fraction_M16.get('ge76')
        0.07745

        >>> isopy.refval.isotope.best_measurement_fraction_M16.get('pd108/pd105')
        1.1849529780564263
        """
        return core.RefValDict(load_refval_values('isotope_best_measurement_fraction_M16'),
                               default_value=np.nan, readonly = True, ratio_function=np.divide,
                               molecule_functions='fraction')

    @core.cached_property
    def initial_solar_system_fraction_L09(self):
        """
        Dictionary containing the isotope fraction of the inital solar system composition from Lodders et al. (2019).

        Reference: `Lodders et al. (2019) <http://materials.springer.com/lb/docs/sm_lbs_978-3-540-88055-4_34>`_.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.

        Examples
        --------
        >>> isopy.refval.isotope.initial_solar_system_fraction_L09['pd105']
        0.2233

        >>> isopy.refval.isotope.initial_solar_system_fraction_L09.get('ge76')
        0.0.07444

        >>> isopy.refval.isotope.initial_solar_system_fraction_L09.get('pd108/pd105')
        1.1849529780564263
        """
        return core.RefValDict(load_refval_values('isotope_initial_solar_system_fraction_L09'),
                               default_value = np.nan, readonly = True, ratio_function=np.divide,
                               molecule_functions='fraction')

    @core.cached_property
    def initial_solar_system_abundance_L09(self):
        """
        Dictionary containing the isotope abundance of the inital solar system composition (normalized to N(Si) = 10^6 atoms) from Lodders et al. (2019).

        Reference: `Lodders et al. (2019) <http://materials.springer.com/lb/docs/sm_lbs_978-3-540-88055-4_34>`_.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.

        Examples
        --------
        >>> isopy.refval.isotope.initial_solar_system_abundance_L09['pd105']
        0.3032

        >>> isopy.refval.isotope.initial_solar_system_abundance_L09.get('ge76')
        8.5

        >>> isopy.refval.isotope.initial_solar_system_abundance_L09.get('pd108/pd105')
        1.184036939313984
        """
        return core.RefValDict(load_refval_values('isotope_initial_solar_system_abundance_L09'),
                               default_value = np.nan, readonly = True, ratio_function=np.divide,
                               molecule_functions='abundance')

    @core.cached_property
    def initial_solar_system_abundance_L09b(self):
        """
        Dictionary containing the isotope abundance calcualted from the elemental abundance and the isotope fraction.

        **Note** These values are not directly taken from the table but calculated from
        *element.initial_solar_system_abundance_L09* and *isotope.initial_solar_system_fraction_L09*. This ensures
        consistency between all three reference values. Discrepancies between the calculated these isotope abundances
        and *isotope.initial_solar_system_abundance_L09* are due to rounding errors.

        Reference: `Lodders et al. (2019) <http://materials.springer.com/lb/docs/sm_lbs_978-3-540-88055-4_34>`_.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.

        Examples
        --------
        >>> isopy.refval.isotope.initial_solar_system_abundance_L09['pd105']
        0.303688

        >>> isopy.refval.isotope.initial_solar_system_abundance_L09.get('ge76')
        8.5606

        >>> isopy.refval.isotope.initial_solar_system_abundance_L09.get('pd108/pd105')
        1.1849529780564263
        """
        element_abundance = isopy.refval.element.initial_solar_system_abundance_L09
        isotope_fraction = isopy.refval.isotope.initial_solar_system_fraction_L09
        result = {k: v * element_abundance.get(k.element_symbol) for k, v in isotope_fraction.items()}

        return core.RefValDict(result, default_value=np.nan, readonly=True, ratio_function=np.divide,
                               molecule_functions='abundance')

    @core.cached_property
    def present_solar_system_fraction_AG89(self):
        """
        Dictionary containing the isotope fraction of the present solar system composition from Anders & Grevesse 1989.

        Reference: `Anders & Grevesse (1989) <https://doi.org/10.1016/0016-7037(89)90286-X>`_.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.
        """
        return core.RefValDict(load_refval_values('isotope_present_solar_system_fraction_AG89'),
                               default_value=np.nan, readonly=True, ratio_function=np.divide,
                               molecule_functions='fraction')

    @core.cached_property
    def initial_solar_system_abundance_AG89(self):
        """
        Dictionary containing the isotope abundance of the initial solar system abundance from Anders & Grevesse 1989.

        Data normalised such that normalized to N(Si) = 10^6 atoms.

        Reference: `Anders & Grevesse (1989) <https://doi.org/10.1016/0016-7037(89)90286-X>`_.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.
        """
        return core.RefValDict(load_refval_values('isotope_initial_solar_system_abundance_AG89'),
                               default_value=np.nan, readonly=True, ratio_function=np.divide,
                               molecule_functions='abundance')

    @core.cached_property
    def present_solar_system_abundance_AG89(self):
        """
        Dictionary containing the isotope abundance of the present solar system abundance from Anders & Grevesse 1989.

        Data normalised such that normalized to N(Si) = 10^6 atoms.

        Reference: `Anders & Grevesse (1989) <https://doi.org/10.1016/0016-7037(89)90286-X>`_.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.
        """
        return core.RefValDict(load_refval_values('isotope_present_solar_system_abundance_AG89'),
                               default_value=np.nan, readonly=True, ratio_function=np.divide,
                               molecule_functions='abundance')

    @core.cached_property
    def sprocess_fraction_B11(self):
        """
        Dictionary containing the estimated s-process fraction of each isotope from Bisterzo et al. (2011).

        Reference: `Bisterzo et al. (2011) <https://doi.org/10.1111/j.1365-2966.2011.19484.x>`_.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.

        Examples
        --------
        >>> isopy.refval.isotope.initial_solar_system_abundance_L09['pd105']
        0.157

        >>> isopy.refval.isotope.initial_solar_system_abundance_L09.get('mo95')
        0.696

        >>> isopy.refval.isotope.initial_solar_system_abundance_L09.get('pd108/pd105')
        4.751592356687898
        """
        return core.RefValDict(load_refval_values('isotope_sprocess_fraction_B11'),
                               ratio_function=np.divide, default_value=np.nan, readonly = True)

class ReferenceValues:
    """ Reference values useful for working with geochemical data"""
    def __init__(self):
        self.reset()

    def __repr__(self):
        out = f"{self.__class__.__name__}("
        space = ' ' * len(out)
        mass = f"""mass=['{"', '".join(self.mass.ls())}']"""
        element = f"""element=['{"', '".join(self.element.ls())}']"""
        isotope = f"""isotope=['{"', '".join(self.isotope.ls())}']"""

        return f'{out}{mass},\n{space}{element},\n{space}{isotope})'

    def _repr_markdown_(self):
        return f'{self.mass._repr_markdown_()}\n\n{self.element._repr_markdown_()}\n\n{self.isotope._repr_markdown_()}'

    def ls(self):
        names = []
        names += [f'mass.{n}' for n in self.mass.ls()]
        names += [f'element.{n}' for n in self.element.ls()]
        names += [f'isotope.{n}' for n in self.isotope.ls()]

        return names

    def __call__(self, path_or_dict, datatype=dict):
        if type(path_or_dict) is str:
            try:
                group, value = path_or_dict.split('.', 1)
            except ValueError:
                raise ValueError(f'Unable to parse "{path_or_dict}"')

            group = getattr(self, group, None)
            if group is None:
                raise ValueError(f'Reference value group "{group}" not found')

            path_or_dict = getattr(group, value, None)
            if path_or_dict is None:
                raise ValueError(f'Reference value "{group.__class__.__name__}.{value}" not found')

        if isinstance(path_or_dict, datatype):
            return path_or_dict
        else:
            return datatype(path_or_dict)

    def reset(self):
        """Resets the reference values to the default values"""
        self._element = element(self)
        self._isotope = isotope(self)
        self._mass = mass(self)

    @property
    def isotope(self):
        """Reference values for different isotopes"""
        return self._isotope

    @property
    def element(self):
        """Reference values for different elements"""
        return self._element

    @property
    def mass(self):
        """Reference values for different atomic masses"""
        return self._mass



import functools
import os as _os
import warnings

from isopy import core
import csv as csv
import datetime as dt
import numpy as np
import chardet
import openpyxl
import pyperclip
from openpyxl import load_workbook
import itertools
import io
import numpy as np


__all__ = ['read_exp',
           'read_csv', 'write_csv',
           'read_xlsx', 'write_xlsx',
           'read_clipboard', 'write_clipboard']

import isopy.checks

NAN_STRINGS = 'nan #NA #N/A N/A NA =NA() =na()'.split()

def rows_to_data(data, has_keys, keys_in_first):
    if len(data) == 0 or len(data[0])==0:
        if has_keys is not True and keys_in_first is None:
            return [[]]
        else:
            return dict()

    # Everything needs to be converted to nan before it gets here
    if has_keys is not False and keys_in_first is None:
        r = [type(v) is str for v in data[0]]
        c = [type(d[0]) is str for d in data]

        if False not in c and False not in r:
            # Both are strings
            try:
                float(data[0][0])
            except ValueError:
                f = True
            else:
                f = False

            try:
                [float(v) for v in data[0][1:]]
            except ValueError:
                r = True
            else:
                r = False

            try:
                [float(v[0]) for v in data[1:]]
            except ValueError:
                c = True
            else:
                c = False

            if has_keys is None and not c and not r:
                if f and (len(data)) == 1:
                    keys_in_first = 'c'
                elif f and len(data[0]) == 1:
                    keys_in_first = 'r'
                elif f:
                    # Both the first row and the first
                    warnings.warn('Both the first row and the first column contain strings. Using first row as keys')
                    keys_in_first = 'r'
                    #raise ValueError('Unable to determine whether the first rows or columns contains the keys')
                else:
                    # All values in the first row and the first column can be converted to float
                    has_keys = False
            elif c and not r:
                keys_in_first = 'c'
            elif r and not c:
                keys_in_first = 'r'
            else:
                # Both the first row and the first column contains at least one value that cant be converted to float
                warnings.warn('Both the first row and the first column contain strings. Using first row as keys')
                keys_in_first = 'r'
                #raise ValueError('Unable to determine whether the first rows or columns contains the keys')

        elif False not in c:
            # Only c contains string
            keys_in_first = 'c'
        elif False not in r:
            # Only r contains strings
            keys_in_first = 'r'
        else:
            # Neither contain only strings
            has_keys=False

    if has_keys is False:
        return data
    elif keys_in_first == 'c':
        return {v[0]: v[1:] for v in data}
    elif keys_in_first == 'r':
        return {v[0]: v[1:] for v in zip(*data)}
    else:
        raise ValueError(f'Unknown value for "keys_in_first" {keys_in_first}')

def data_to_rows(data, keys_in_first, keyfmt = None):
    data = isopy.asanyarray(data)

    if isinstance(data, core.IsopyArray):
        if data.ndim == 0:
            data = data.reshape(-1)
        if keys_in_first == 'r':
            rows = [[k.str(keyfmt) for k in data.keys()]]
            rows += data.to_list()
        elif keys_in_first == 'c':
            rows = [[k.str(keyfmt)] + v.tolist() for k, v in data.items()]
        else:
            raise ValueError(f'Unknown value for "keys_in_first" {keys_in_first}')
    else:
        if data.ndim == 0:
            data = data.reshape((1, 1))
        elif data.ndim == 1:
            data = data.reshape((1, -1))
        elif data.ndim > 2:
            raise ValueError('data cannot have more than 2 dimensions')
        rows = data.tolist()

    return rows

################
### read exp ###
################
class NeptuneData:
    """
    Container for the data returned by ``read_exp``.
    """
    def __init__(self, info, cycle, time, measurements):
        self.info = info
        self.cycle = cycle
        self.time = time
        self.measurements = measurements


def read_exp(filename, rename = None) -> NeptuneData:
    """
    Load data from a Neptune/Triton export file.

    Parameters
    ----------
    filename : str, bytes, StringIO, BytesIO
        Path for file to be opened. Alternatively a file like byte string or a file like object can be supplied.
    rename : dict, Callable, Optional
        For renaming keys in the analysed data. Useful for cases when the key is the mass rather
        than the isotope measured. If a dictionary is passed then every key present in the
        dictionary will be replaced by the associated value. A callable can also be passed that
        takes the key in the file and returns the new key.

    Returns
    -------
    neptune_data : NeptuneData
        An object containing the following attributes:

        * info - Dictionary containing the metadata included at the beginning of the file.
        * cycle - A list containing the cycle number for each measurement.
        * time - A list containing datetime objects for each measurement.
        * measurements - An dictionary containing an an isopy array with the values in for each line measured. Static measurements are always given as line ``1``.

         To extract e.g only the isotope data from a measurement use ``neptune_data.measurement[line].copy(flavour_eq='isotope')``.
    """


    # If filename is a string load the files.
    if type(filename) is str:
        with open(filename, 'rb') as fileio:
            file = fileio.read()
    elif type(filename) is bytes:
        file = filename
    elif type(filename) is io.BytesIO:
        filename.seek(0)
        file = filename.read()
    elif type(filename) is io.StringIO:
        filename.seek(0)
        file = filename.read()
    else:
        raise TypeError('filename is of unknown type')

    if type(file) is bytes:
        # find the files encoding
        encoding = chardet.detect(file).get('encoding')

        # Decode the bytes into string.
        file = file.decode(encoding)

    csv_reader = csv.reader(io.StringIO(file), dialect='excel-tab')

    information = {}
    for row in csv_reader:
        if ':' in row[0]:  # meta
            name, value = row[0].split(':', 1)
            information[name.strip()] = value.strip()

        if row[0] == 'Cycle':
            # We can reuse the function from before to read the integration data
            data = _read_csv_data(row, csv_reader, termination_symbol='***')
            data = rows_to_data(data, True, 'r')

    if rename is None:
        renamer = lambda name: name
    elif isinstance(rename, dict):
        renamer = lambda name: rename.get(name, name)
    elif callable(rename):
        renamer = rename
    else:
        raise TypeError('rename must be a dict or a callable function')

    data.pop("", None)
    cycle = [int(i) for i in data.pop('Cycle')]

    time = data.pop('Time')
    try:
        time = [dt.datetime.strptime(time[i], '%H:%M:%S:%f') for i in range(len(time))]
    except:
        # Dont think the time will ever be given as a float
        # time = [dt.datetime.fromtimestamp(time[i]).strftime('%H:%M:%S:%f') for i in range(len(time))]
        warnings.warn('Unable to parse the time column')
        time = [dt.datetime.fromtimestamp(0).strftime('%H:%M:%S:%f') for i in range(len(time))]

    measurements = {}

    for key in data:
        # Check if the data has more than one line
        if ":" in key:
            line, keystring = key.split(":", 1)
            line = int(line)
        else:
            line = 1
            keystring = key

        # Rename columns if needed
        keystring = renamer(keystring)
        measurements.setdefault(line, {})[isopy.keystring(keystring)] = data[key]

    # Convert to isopy arrays
    measurements = {line: isopy.asarray(measurements[line]) for line in measurements}

    return NeptuneData(information, cycle, time, measurements)

######################
### read/write CSV ###
######################
def read_csv(filename, has_keys= None, keys_in_first = None, comment_symbol ='#', encoding = None, dialect = 'excel'):
    """
    Load data from a csv file.

    Parameters
    ----------
    filename : str, bytes, StringIO, BytesIO
        Path for file to be opened. Alternatively a file like byte string can be supplied.
        Also accepts file like objects.
    has_keys : bool, None
        If True or *keys_in_first* is not None a dictionary will always be returned. If False an nexted list is
        always returned. If None it will return a nested list of all values in the first row and column
        can be converted to/is a float.
    keys_in_first : {'c', 'r', None}
        Where the keys are found. Give 'r' if the keys are found in the first row and 'c' if the
        keys are in first column. If None it will analyse the data to determine where the keys are. If *has_keys*
        is not False an exception will be raised if it cant determine where the keys are.
    comment_symbol : str, Default = '#'
        Rows starting with this string will be ignored.
    encoding : str
        Encoding of the file. If None the encoding will be guessed from the file.
    dialect
        Dialect of the csv file. If None the dialect will be guessed from the file.

    Returns
    -------
    data : dict or list
        Returns a dictionary for data with keys otherwise a list.
    """

    #If filename is a string load the files.
    if type(filename) is str:
        with open(filename, 'rb') as fileio:
            text = fileio.read()
    elif type(filename) is bytes:
        text = filename
    elif type(filename) is io.BytesIO:
        filename.seek(0)
        text = filename.read()
    elif type(filename) is io.StringIO:
        filename.seek(0)
        text = filename.read()
    else:
        raise TypeError(f'filename is of unknown type {type(filename)}')

    if type(text) is bytes:
        #find the files encoding
        if encoding is None:
            encoding = chardet.detect(text).get('encoding')

        #Decode the bytes into string.
        text = text.decode(encoding)

    #find the csv dialect
    if dialect is None:
        dialect = csv.Sniffer().sniff(text)

    # Create a reader object by converting the file string to a file-like object
    csv_reader = csv.reader(io.StringIO(text), dialect=dialect)
    for row in csv_reader:
        row_data = [r.strip() for r in row]  # Remove any training whitespaces from the data in this row
        if len(row_data) == 0:
            return rows_to_data([[]], has_keys, keys_in_first)

        if comment_symbol is not None and row[0][:len(comment_symbol)] == comment_symbol:
            pass # This is a comment so ignore it.
        else:
            data = _read_csv_data(row_data, csv_reader, comment_symbol)
            return rows_to_data(data, has_keys, keys_in_first)

    return rows_to_data([[]], has_keys, keys_in_first)

def _read_csv_data(first_row, reader, comment_symbol=None, termination_symbol=None):
    data = []
    dlen = None

    # Loop over the remaining rows
    for i, row in enumerate(itertools.chain([first_row], reader)):
        row = [r.strip() for r in row]
        if len(row) > 0:
            if termination_symbol is not None and row[0][:len(termination_symbol)] == termination_symbol:
                # Stop reading data if we find this string at the beginning of a row
                break

            if comment_symbol is not None and row[0][:len(comment_symbol)] == comment_symbol and row[0] not in NAN_STRINGS:
                # Row is a comment, ignore
                continue

        data.append([])
        if dlen is None:
            dlen = len(row)

        elif dlen != len(row):
            raise ValueError('Rows have different number of values')

        for j in range(len(row)):
            value = row[j]

            if value in NAN_STRINGS:
                value = float('nan')

            data[-1].append(value)

    return data


def write_csv(filename, data, comments=None, keys_in_first='r', comment_symbol = '#', keyfmt = None,
              dialect = 'excel') -> None:
    """
    Save data to a csv file.

    Parameters
    ----------
    filename : str, StringIO, BytesIO
        Path/name of the csv file to be created. Any existing file with the same path/name will be
        over written. Also accepts file like objects.
    data : isopy_array_like, numpy_array_like
        Data to be saved in the array.
    comments : str, Sequence[str], Optional
        Comments to be included at the top of the file
    keys_in_first : {'c', 'r'}
        Only used if the input has keys. Give 'r' if the keys should be in the first row and 'c' if the
        keys should be in the first column.
    comment_symbol : str, Default = '#'
        This string will precede any comments at the beginning of the file.
    keyfmt
        Specify the format used for the key string. See the ``str()`` method of each key string for options.
    dialect
        The CSV dialect used to save the file. Default to 'excel' which is a ', ' seperated file.
    """

    rows = data_to_rows(data, keys_in_first, keyfmt)

    if comments is not None:
        if type(comments) is not list:
            comments = [comments]

        crows = []
        for comment in comments:
            crows.append([f'{comment_symbol}{comment}'] + ['' for i in range(len(rows[0][1:]))])

        if len(rows[0]) == 0:
            rows = crows
        else:
            rows = crows + rows

    if type(filename) is str:
        with open(filename, mode='w', newline='') as file:
            csv_writer = csv.writer(file, dialect=dialect)
            for row in rows:
                csv_writer.writerow(row)

    elif type(filename) is io.StringIO:
        filename.truncate(0)
        filename.seek(0)
        csv_writer = csv.writer(filename, dialect=dialect)
        for row in rows:
            csv_writer.writerow(row)

    elif type(filename) is io.BytesIO:
        filename.truncate(0)
        filename.seek(0)

        file = io.StringIO()
        csv_writer = csv.writer(file, dialect=dialect)
        for row in rows:
            csv_writer.writerow(row)
        file.seek(0)
        text = file.read()
        filename.write(text.encode('UTF-8'))
    else:
        raise TypeError(f'filename is of unknown type {type(filename)}')

def read_clipboard(has_keys= None, keys_in_first = None, comment_symbol ='#', dialect = None):
    """
    Load data from values in the clipboard.

    Parameters
    ----------
    comment_symbol : str, Default = '#'
        Rows starting with this string will be ignored.
    has_keys : bool, None
        If True or *keys_in_first* is not None a dictionary will always be returned. If False an nexted list is
        always returned. If None it will return a nested list of all values in the first row and column
        can be converted to/is a float.
    keys_in_first : {'c', 'r', None}
        Where the keys are found. Give 'r' if the keys are found in the first row and 'c' if the
        keys are in first column. If None it will analyse the data to determine where the keys are. If *has_keys*
        is not False an exception will be raised if it cant determine where the keys are.
    dialect
        Dialect of the values in the clipboard. If None the dialect will be guessed from the values.
    """
    data = pyperclip.paste()
    return read_csv(io.StringIO(data), comment_symbol=comment_symbol, has_keys=has_keys,
                    keys_in_first=keys_in_first, dialect=dialect)

def write_clipboard(data, comments=None, keys_in_first='r',
              comment_symbol = '#', keyfmt = None, dialect = 'excel'):
    """
    Copies data to the clipboard

    Parameters
    ----------
    data : isopy_array_like, numpy_array_like
        Data to be copied.
    comments : str, Sequence[str], Optional
        Comments to be included
    keys_in_first : {'c', 'r'}
        Only used if the input has keys. Give 'r' if the keys should be in the first row and 'c' if the
        keys should be in the first column.
    comment_symbol : str, Default = '#'
        This string will precede any comments.
    keyfmt
        Specify the format used for the key string. See the ``str()`` method of each key string for options.
    dialect
        The CSV dialect used to copy the data to the clipboard. Default to 'excel' which is gives ', ' seperated data.
    """
    text = io.StringIO()
    write_csv(text, data, comments=comments, keys_in_first=keys_in_first, dialect=dialect,
              comment_symbol=comment_symbol, keyfmt=keyfmt)

    text.seek(0)
    pyperclip.copy(text.read())

#######################
### Read/write xlsx ###
#######################
def read_xlsx(filename, sheetname=None, has_keys = None, keys_in_first=None, comment_symbol='#', start_at='A1'):
    """
    Load data from an excel file.

    Parameters
    ----------
    filename : str, bytes, BytesIO
        Path for file to be opened. Also accepts file like objects.
    sheetname : str, int, Optional
        To load data from a single sheet in the workbook pass either the name of the sheet or
        the position of the sheet. If nothing is specified all the data for all sheets in the
        workbook is returned.
    has_keys : bool, None
        If True or *keys_in_first* is not None a dictionary will always be returned. If False an nexted list is
        always returned. If None it will return a nested list of all values in the first row and column
        can be converted to/is a float.
    keys_in_first : {'c', 'r', None}
        Where the keys are found. Give 'r' if the keys are found in the first row and 'c' if the
        keys are in first column. If None it will analyse the data to determine where the keys are. If *has_keys*
        is not False an exception will be raised if it cant determine where the keys are.
    comment_symbol : str, Default = '#'
        Rows starting with this string will be ignored.
    start_at : str, (int, int)
        Start scanning at this cell. Can either be a excel style cell reference or a (row, column) tuple of integers.

    Returns
    -------
    data : dict
        A dictionary containing the data for a single sheet or a dictionary containing the data
        for multiple sheets in the workbook indexed by sheet name.
    """
    if type(filename) is bytes:
        filename = io.BytesIO(filename)

    if type(filename) is openpyxl.Workbook:
        workbook = filename
    else:
        workbook = openpyxl.load_workbook(filename, data_only=True, read_only=True)

    sheetnames = workbook.sheetnames

    if sheetname is None:
        sheet_data = {}
        for sheetname in sheetnames:
            sheet_data[sheetname] = _read_xlsx_sheet(workbook[sheetname], has_keys, keys_in_first, comment_symbol, start_at)
        return sheet_data

    elif type(sheetname) is int and sheetname < len(sheetnames):
        return _read_xlsx_sheet(workbook[sheetnames[sheetname]], has_keys, keys_in_first, comment_symbol, start_at)

    elif sheetname in sheetnames:
        return _read_xlsx_sheet(workbook[sheetname], has_keys, keys_in_first, comment_symbol, start_at)

    else:
        raise ValueError(f'sheetname "{sheetname}" not found in workbook"')


def _read_xlsx_sheet(worksheet, has_keys, keys_in_first, comment_symbol, start_at):
    if type(start_at) is tuple:
        start_r, start_c = start_at
    else:
        start_r, start_c = openpyxl.utils.cell.coordinate_to_tuple(start_at)

    #Remove inital comments
    for ri in range(start_r, worksheet.max_row + 1):
        value = worksheet.cell(ri, start_c).value
        if type(value) is str and value[:len(comment_symbol)] == comment_symbol and value not in NAN_STRINGS:
            continue
        else:
            start_r = ri
            break

    # Select area containing data
    stop_c = worksheet.max_column + 1
    stop_r = worksheet.max_row + 1

    for ri in range(start_r + 1, stop_r+1):
        values_r = []
        for ci in range(start_c, stop_c):
            if ri != stop_r:
                # All rows should be included
                cell = worksheet.cell(ri, start_c)
                value = cell.value

                if value == '':
                    values_r.append(None)
                else:
                    values_r.append(value)

            if values_r.count(None) == len(values_r):
                values_c = []
                if ci != (stop_c-1):
                    for rj in range(start_r, ri + 1):
                        cell = worksheet.cell(rj, ci + 1)
                        value = cell.value

                        if value == '':
                            values_c.append(None)
                        else:
                            values_c.append(value)

                if values_c.count(None) == len(values_c):
                    stop_r = ri
                    stop_c = ci + 1

            if ci == stop_c: break
        if ri == stop_r: break

    if (stop_r-start_r) == 1 and (stop_c-start_c) == 1 and ((cell:=worksheet.cell(start_r,start_c)).value is None or cell.value==''):
        return rows_to_data([[]], has_keys, keys_in_first)

    data = []

    # Iterate over each column
    for ri in range(start_r, stop_r):
        value = worksheet.cell(ri, start_c).value
        if type(value) is str and value[:len(comment_symbol)] == comment_symbol and value not in NAN_STRINGS:
            continue

        data.append([])
        for ci in range(start_c, stop_c):
            cell = worksheet.cell(ri, ci)
            value = cell.value
            if type(value) is str:
                value = value.strip()

            if cell.data_type == 'e' or (type(value) is str and value.upper() == '=NA()'):
                data[-1].append(float('nan'))
            elif value == '' or value in NAN_STRINGS or value is None:
                data[-1].append(float('nan'))
            else:
                data[-1].append(value)

    return rows_to_data(data, has_keys, keys_in_first)


def write_xlsx(filename, *sheets, comments = None,
               keys_in_first= 'r', comment_symbol= '#', keyfmt = None,
               start_at ="A1", append = False, clear = True, **sheetnames):
    """
    Save data to an excel file.

    Parameters
    ----------
    filename : str, BytesIO
        Path/name of the excel file to be created. Any existing file with the same path/name
        will be overwritten. Also accepts file like objects.
    sheets : isopy_array_like, numpy_array_like
        Data given here will be saved as sheet1, sheet2 etc.
    comments : str, Sequence[str], Optional
        Comments to be included at the top of the file
    keys_in_first : {'c', 'r'}
        Only used if the input has keys. Give 'r' if the keys should be in the first row and 'c' if the
        keys should be in the first column.
    comment_symbol : str, Default = '#'
        This string will precede any comments at the beginning of the file
    keyfmt
        Specify the format used for the key string. See the ``str()`` method of each key string for options.
    start_at: str, (int, int)
        The first cell where the data is written. Can either be a excel style cell reference or a (row, column)
        tuple of integers.
    append : bool, Default = False
        If ``True`` and *filename* exists it will append the data to this workbook. An exception
        is raised if *filename* is not a valid excel workbook.
    clear : bool, Default = True
        If ``True`` any preexisting sheets are cleared before any new data is written to it.
    sheetnames : isopy_array, numpy_array
        Data given here will be saved in a sheet with *sheetname* name.
    """
    save = True
    if type(filename) is openpyxl.Workbook:
        workbook = filename
        save = False
    elif type(filename) is io.BytesIO:
        if append and filename.seek(0,2) > 0:
            workbook = openpyxl.load_workbook(filename=filename)
        else:
            # openpyxl truncates BytesIO objects automatically via the use of ZipFile('w')

            workbook = openpyxl.Workbook()
            workbook.remove(workbook.active)
    elif type(filename) is str:
        if append and _os.path.exists(filename):
            workbook = openpyxl.load_workbook(filename=filename)
        else:
            workbook = openpyxl.Workbook()
            workbook.remove(workbook.active)

    sheetname_data = {f'sheet{i+1}': data for i, data in enumerate(sheets)}
    sheetname_data.update(sheetnames)
    try:
        for sheetname, data in sheetname_data.items():
            #If appending then delete any preexisting workbooks
            if sheetname not in workbook.sheetnames:
                worksheet = workbook.create_sheet(sheetname)
            elif clear:
                workbook.remove(workbook[sheetname])
                worksheet = workbook.create_sheet(sheetname)
            else:
               worksheet = workbook[sheetname]

            _write_xlsx(worksheet, data, comments, comment_symbol, keys_in_first, start_at, keyfmt)

        #Workbooks must have at least one sheet
        if len(workbook.sheetnames) == 0:
            workbook.create_sheet('sheet1')

        if save: workbook.save(filename)
    finally:
        if save: workbook.close()

def _write_xlsx(worksheet, data, comments, comment_symbol, keys_in_first, start_at, keyfmt):
    rows = data_to_rows(data, keys_in_first, keyfmt)
    if type(start_at) is tuple:
        start_r, start_c = start_at
    else:
        start_r, start_c = openpyxl.utils.cell.coordinate_to_tuple(start_at)

    if comments is not None:
        if not isinstance(comments, (list, tuple)):
            comments = [comments]
        for comment in comments:
            worksheet.cell(start_r, start_c).value = f'{comment_symbol}{comment}'
            start_r += 1

    for ri, row in enumerate(rows):
        for ci, value in enumerate(row):
            if str(value) == 'nan':
                value = '#N/A'
            if value is not None and value != '':
                worksheet.cell(start_r + ri, start_c + ci).value = value



import numpy as np
#from numpy.lib.function_base import array_function_dispatch
from scipy import stats
import functools

import isopy
from . import core
import warnings

__all__ = ['sd', 'nansd', 'se', 'nanse', 'mad', 'nanmad', 'nancount',
           'rstack', 'cstack', 'concatenate',
           'arrayfunc', 'keymax', 'keymin',
           'add', 'subtract', 'power', 'multiply', 'divide',
           'is_outlier', 'not_outlier', 'upper_limit', 'lower_limit',
           'approved_numpy_functions']

__all__ += 'sd2 sd3 sd95 sd99 nansd2 nansd3 nansd95 nansd99'.split()
__all__ += 'se2 se3 se95 se99 nanse2 nanse3 nanse95 nanse99'.split()
__all__ += 'mad2 mad3 mad95 mad99 nanmad2 nanmad3 nanmad95 nanmad99'.split()

_min = min
_max = max
def calculate_ci(ci, df=None):
    if df is None:
        return stats.norm.ppf(0.5 + ci / 2)
    else:
        return stats.t.ppf(0.5 + ci / 2, df)

def arrayfunc_wrapper(func):
    @functools.wraps(func)
    def call_arrayfunc(*args, **kwargs):
        return arrayfunc(func, *args, **kwargs)
    return call_arrayfunc

def arrayfunc(func, *inputs, keys=None, **kwargs):
    """
    arrayfunc(func, *inputs : isopy_array, dict, keys=None : keystring_like, Optional, **kwargs) -> IsopyArray
    arrayfunc(func, *inputs : isopy_array, scalar, keys=None : keystring_like, Optional, **kwargs) -> IsopyArray
    arrayfunc(func, *inputs : isopy_array, keys=None : keystring_like, Optional, **kwargs) -> IsopyArray
    arrayfunc(func, *inputs : dict, scalar, keys=None : keystring_like, Optional, **kwargs) -> RefValDict
    arrayfunc(func, *inputs : dict, keys=None : keystring_like, Optional, **kwargs) -> RefValDict
    arrayfunc(func, *inputs : scalar, keys=None : keystring_like, **kwargs) -> IsopyArray
    arrayfunc(func, *inputs : scalar, **kwargs) -> scalar

    Call a function *func* on each column in the *inputs*.

    If the input consists of isopy arrays and dictionaries then the returned array will only contain the keys of
    the arrays.

    If the input consists of more than one isopy array the returned array will contain the keys of both arrays in
    sorted order.

    If *kwargs* contain any key filters these will be applied the input values before the function is computed.

    Parameters
    ----------
    func
        a function to be applied the to input.
    *inputs
        The input for the function.
    keys
        If given the returned output will contain these keys, irregardless of the keys of the input.
    **kwargs:
        Key word arguments to be used with *func*. If *kwargs* contain any key filters these will be applied
        the input values before the function is computed.

    Examples
    --------
    >>> array = isopy.random(100, keys=('ru', 'pd', 'cd')
    >>> isopy.arrayfunc(np.std, array, keys=('ru', 'cd'))
    (row)      Ru (f8)    Cd (f8)
    -------  ---------  ---------
    None       1.12864    0.97975
    IsopyNdarray(-1, flavour='element', default_value=nan)

    >>> isopy.arrayfunc(np.std, array, keys_eq=('ru', 'cd'))
    (row)      Ru (f8)    Cd (f8)
    -------  ---------  ---------
    None       1.12864    0.97975
    IsopyNdarray(-1, flavour='element', default_value=nan)

    >>> isopy.arrayfunc(np.add, 2, 3, keys=('ru', 'cd'))
    (row)      Ru (f8)    Cd (f8)
    -------  ---------  ---------
    None       5.00000    5.00000
    IsopyNdarray(-1, flavour='element', default_value=nan)
    """
    if kwargs:
        key_filters = {key: kwargs.pop(key) for key in tuple(kwargs.keys()) if
                       key.rsplit('_')[-1] in core.KEY_COMPARISONS}
    else:
        key_filters = None

    return core.call_array_function(func, *inputs, keys=keys, key_filters=key_filters, **kwargs)


#@array_function_dispatch(_sd_dispatcher)
@arrayfunc_wrapper
def sd(a, axis = None, *, ci = None, zscore=None): #, where = core.NotGiven):
    """
    Compute the standard deviation along the specified axis for N-1 degrees of freedom.

    Shortcut for ``np.std(a, ddof = 1, axis=axis)``.

    If *ci* is given the result will be multiplied by the confidence interval *ci* of a
    t-distribution with N-1 degrees of freedom.

    If *zscore* is given then the result is multiplied by *zscore*.

    Functions ``sd2``, ``sd3``, ``sd95`` ``sd99`` are also avaliable for z-score
    values of 2 and 3 and for 95% and 99% confidence intervals respectivley.

    Examples
    --------
    >>> array = isopy.array(ru = [np.nan, 1.1, 2.2, 1.8],
                            pd = [3.1, 3.8, 2.9, 4.2],
                            cd = [6.1, 5.8, 4.7, 8.1])
    >>> isopy.sd(array)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None           nan    0.60553    1.41745
    IsopyNdarray(-1, flavour='element', default_value=nan)
    >>> isopy.sd(array, axis=1)
    array([       nan, 2.35867194, 1.28970281, 3.17962262])
    >>> isopy.sd2(array) #same as sd(array, zscore=2)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None           nan    1.21106    2.83490
    IsopyNdarray(-1, flavour='element', default_value=nan)
    >>> isopy.sd95(array) #same as sd(array, ci=0.95)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None           nan    1.92707    4.51096
    IsopyNdarray(-1, flavour='element', default_value=nan)

    See Also
    --------
    :func:`nansd`, `np.std <https://numpy.org/devdocs/reference/generated/numpy.std.html#numpy.std>`_
    """
    result = np.std(a, ddof=1, axis=axis)
    if ci is not None:
        df = np.count_nonzero(np.invert(np.isnan(a)), axis=axis) - 1
        return result * calculate_ci(ci, df)

    if zscore is not None:
        return result * zscore

    return result


#@array_function_dispatch(_sd_dispatcher)
@arrayfunc_wrapper
def nansd(a, axis = None, *, ci = None, zscore=None): #, where = core.NotGiven):
    """
    Compute the standard deviation along the specified axis for N-1 degrees of freedom, while ignoring NaNs.

    Shortcut for ``np.nanstd(a, ddof = 1, axis=axis)``.

    If *ci* is given the result will be multiplied by the confidence interval *ci* of a
    t-distribution with N-1 degrees of freedom.

    If *zscore* is given then the result is multiplied by *zscore*.

    Functions ``nansd2``, ``nansd3``, ``nansd95`` ``nansd99`` are also avaliable for z-score
    values of 2 and 3 and for 95% and 99% confidence intervals respectivley.


    Examples
    --------
    >>> array = isopy.array(ru = [np.nan, 1.1, 2.2, 1.8],
                            pd = [3.1, 3.8, 2.9, 4.2],
                            cd = [6.1, 5.8, 4.7, 8.1])
    >>> isopy.nansd(array)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None       0.55678    0.60553    1.41745
    IsopyNdarray(-1, flavour='element', default_value=nan)
    >>> isopy.nansd(array, axis=1)
    array([2.12132034, 2.35867194, 1.28970281, 3.17962262])
    >>> isopy.nansd2(array) #same as nansd(array, zscore=2)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None       1.11355    1.21106    2.83490
    IsopyNdarray(-1, flavour='element', default_value=nan)
    >>> isopy.nansd95(array) #same as nansd(array, ci=0.95)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None       2.39562    1.92707    4.51096
    IsopyNdarray(-1, flavour='element', default_value=nan)

    See Also
    --------
    :func:`sd`, `numpy.nanstd <https://numpy.org/devdocs/reference/generated/numpy.nanstd.html#numpy.nanstd>`_
    """
    result = np.nanstd(a, ddof=1, axis=axis)

    if ci is not None:
        df = np.count_nonzero(np.invert(np.isnan(a)), axis=axis) - 1
        return result * calculate_ci(ci, df)

    if zscore is not None:
        return result * zscore

    return result

#@array_function_dispatch(_se_dispatcher)
@arrayfunc_wrapper
def se(a, axis=None, *, ci = None, zscore=None): #, where = core.NotGiven):
    """
    Compute the standard error along the specified axis for N-1 degrees of freedom.

    Shortcut for ``scipy.stats.sem(a, axis=axis, nan_policy='propagate')``.

    If *ci* is given the result will be multiplied by the confidence interval *ci* of a
    t-distribution with N-1 degrees of freedom.

    If *zscore* is given then the result is multiplied by *zscore*.

    Functions ``se2``, ``se3``, ``se95`` ``se99`` are also avaliable for z-score
    values of 2 and 3 and for 95% and 99% confidence intervals respectivley.

    Examples
    --------
    >>> array = isopy.array(ru = [np.nan, 1.1, 2.2, 1.8],
                            pd = [3.1, 3.8, 2.9, 4.2],
                            cd = [6.1, 5.8, 4.7, 8.1])
    >>> isopy.se(array)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None           nan    0.30277    0.70873
    IsopyNdarray(-1, flavour='element', default_value=nan)
    >>> isopy.se(array, axis=1)
    array([       nan, 1.36177988, 0.74461026, 1.83575598])
    >>> isopy.se2(array) #same as se(array, zscore=2)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None           nan    0.60553    1.41745
    IsopyNdarray(-1, flavour='element', default_value=nan)
    >>> isopy.se95(array) #same as se(array, ci=0.95)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None           nan    0.96353    2.25548
    IsopyNdarray(-1, flavour='element', default_value=nan)

    See Also
    --------
    :func:`nanse`, `scipy.stats.sem <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.sem.html?highlight=sem#scipy.stats.sem>`_
    """
    result =  stats.sem(a, nan_policy='propagate', axis=axis)

    if ci is not None:
        df = np.count_nonzero(np.invert(np.isnan(a)), axis=axis) - 1
        return result * calculate_ci(ci, df)

    if zscore is not None:
        return result * zscore

    return result

#@array_function_dispatch(_se_dispatcher)
@arrayfunc_wrapper
def nanse(a, axis=None, *, ci = None, zscore=None): #, where = core.NotGiven):
    """
    Compute the standard error along the specified axis for N-1 degrees of freedom, while ignoring
    NaNs.

    Shortcut for ``scipy.stats.sem(a, axis=axis, nan_policy='omit')``.

    If *ci* is given the result will be multiplied by the confidence interval *ci* of a
    t-distribution with N-1 degrees of freedom.

    If *zscore* is given then the result is multiplied by *zscore*.

    Functions ``nanse2``, ``nanse3``, ``nanse5`` ``nanse99`` are also avaliable for z-score
    values of 2 and 3 and for 95% and 99% confidence intervals respectivley.

    Examples
    --------
    >>> array = isopy.array(ru = [np.nan, 1.1, 2.2, 1.8],
                            pd = [3.1, 3.8, 2.9, 4.2],
                            cd = [6.1, 5.8, 4.7, 8.1])
    >>> isopy.nanse(array)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None       0.32146    0.30277    0.70873
    IsopyNdarray(-1, flavour='element', default_value=nan)
    >>> isopy.nanse(array, axis=1) #returns a masked array
    masked_array(data=[1.4999999999999996, 1.3617798810543664,
                   0.7446102634562892, 1.835755975068582],
             mask=[False, False, False, False],
       fill_value=1e+20)
    >>> isopy.nanse2(array) #same as nanse(array, zscore=2)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None       0.64291    0.60553    1.41745
    IsopyNdarray(-1, flavour='element', default_value=nan)
    >>> isopy.nanse95(array) #same as nanse(array, ci=0.95)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None       1.38311    0.96353    2.25548
    IsopyNdarray(-1, flavour='element', default_value=nan)


    See Also
    --------
    :func:`se`, `scipy.stats.sem <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.sem.html?highlight=sem#scipy.stats.sem>`_
    """
    result = stats.sem(a, nan_policy='omit', axis=axis)

    if ci is not None:
        df = np.count_nonzero(np.invert(np.isnan(a)), axis=axis) - 1
        return result * calculate_ci(ci, df)

    if zscore is not None:
        return result * zscore

    return result

#@array_function_dispatch(_mad_dispatcher)
@arrayfunc_wrapper
def mad(a, axis=None, scale= 'normal', *, ci = None, zscore=None): #,  where = core.NotGiven):
    """
    Compute the median absolute deviation of the data along the given axis.

    Shortcut for ``scipy.stats.median_abs_deviation(a, axis, scale=scale, nan_policy='propagate')``.

    If *ci* is given the result will be multiplied by the confidence interval *ci* of a
    t-distribution with N-1 degrees of freedom.

    If *zscore* is given then the result is multiplied by *zscore*.

    Functions ``mad2``, ``mad3``, ``mad95`` ``mad99`` are also avaliable for z-score
    values of 2 and 3 and for 95% and 99% confidence intervals respectivley.

    Examples
    --------
    >>> array = isopy.array(ru = [np.nan, 1.1, 2.2, 1.8],
                            pd = [3.1, 3.8, 2.9, 4.2],
                            cd = [6.1, 5.8, 4.7, 8.1])
    >>> isopy.mad(array)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None           nan    0.66717    1.03782
    IsopyNdarray(-1, flavour='element', default_value=nan)
    >>> isopy.mad(array, axis=1)
    array([       nan, 2.96520444, 1.03782155, 3.55824532])
    >>> isopy.mad2(array) #same as mad(array, zscore=2)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None           nan    1.33434    2.07564
    IsopyNdarray(-1, flavour='element', default_value=nan)
    >>> isopy.mad95(array) #same as mad(array, ci=0.95)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None           nan    2.12324    3.30281
    IsopyNdarray(-1, flavour='element', default_value=nan)


    See Also
    --------
    :func:`nanmad`, `scipy.stats.median_abs_deviation <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.median_abs_deviation.html#scipy.stats.median_abs_deviation>`_
    """
    # function sometimes return a float value
    result = np.asarray( stats.median_abs_deviation(a, scale=scale, nan_policy='propagate', axis=axis) )

    if ci is not None:
        df = np.count_nonzero(np.invert(np.isnan(a)), axis=axis) - 1
        return result * calculate_ci(ci, df)

    if zscore is not None:
        return result * zscore

    # result is not always an array
    return result

#@array_function_dispatch(_mad_dispatcher)
@arrayfunc_wrapper
def nanmad(a, axis=None, scale = 'normal', *, ci = None, zscore=None): #, where = core.NotGiven):
    """
    Compute the median absolute deviation along the specified axis, while ignoring NaNs.

    Shortcut for ``scipy.stats.median_abs_deviation(a, axis, scale=scale, nan_policy='omit')``.

    If *ci* is given the result will be multiplied by the confidence interval *ci* of a
    t-distribution with N-1 degrees of freedom.

    If *zscore* is given then the result is multiplied by *zscore*.

    Functions ``nanmad2``, ``nanmad3``, ``nanmad95`` ``nanmad99`` are also avaliable for z-score
    values of 2 and 3 and for 95% and 99% confidence intervals respectivley.

    Examples
    --------
    >>> array = isopy.array(ru = [np.nan, 1.1, 2.2, 1.8],
                            pd = [3.1, 3.8, 2.9, 4.2],
                            cd = [6.1, 5.8, 4.7, 8.1])
    >>> isopy.nanmad(array)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None       0.59304    0.66717    1.03782
    IsopyNdarray(-1, flavour='element', default_value=nan)
    >>> isopy.nanmad(array, axis=1)
    array([2.22390333, 2.96520444, 1.03782155, 3.55824532])
    >>> isopy.nanmad2(array) #same as nanmad(array, zscore=2)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None       1.18608    1.33434    2.07564
    IsopyNdarray(-1, flavour='element', default_value=nan)
    >>> isopy.nanmad95(array) #same as nanmad(array, ci=0.95)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None       2.55165    2.12324    3.30281
    IsopyNdarray(-1, flavour='element', default_value=nan)

    See Also
    --------
    :func:`mad`, `scipy.stats.median_abs_deviation <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.median_abs_deviation.html#scipy.stats.median_abs_deviation>`_
    """
    # function sometimes return a float value
    result = np.asarray( stats.median_abs_deviation(a, scale=scale, nan_policy = 'omit', axis=axis) )

    if ci is not None:
        df = np.count_nonzero(np.invert(np.isnan(a)), axis=axis) - 1
        return result * calculate_ci(ci, df)

    if zscore is not None:
        return result * zscore

    return result

#@array_function_dispatch(_count_dispatcher)
@arrayfunc_wrapper
def nancount(a, axis=None): #, where = core.NotGiven):
    """
    Count all values in array that are not NaN or Inf along the specified axis.

    Shortcut for ``np.count_nonzero(np.isfinite(a), axis=axis)``.

    Examples
    --------
    >>> array = isopy.array(ru = [np.nan, 1.1, 2.2, 1.8],
                            pd = [3.1, 3.8, 2.9, 4.2],
                            cd = [6.1, 5.8, 4.7, 8.1])
    >>> isopy.nancount(array)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None       3.00000    4.00000    4.00000
    IsopyNdarray(-1, flavour='element', default_value=nan)
    >>> isopy.nancount(array, axis=1)
    array([2, 3, 3, 3])
    """
    #if where is not core.NotGiven:
    #    a = a[where]
    return np.count_nonzero(np.isfinite(a), axis=axis)

for func in [sd, nansd, se, nanse, mad, nanmad, nancount]:
    core.APPROVED_FUNCTIONS.append(func)

### preset functions
# These dont need to be approved
sd2 = core.partial_func(sd, 'sd2', zscore=2)
sd3 = core.partial_func(sd, 'sd3', zscore=3)
sd95 = core.partial_func(sd, 'sd95', ci=0.95)
sd99 = core.partial_func(sd, 'sd99', ci=0.99)
nansd2 = core.partial_func(nansd, 'nansd2', zscore=2)
nansd3 = core.partial_func(nansd, 'nansd3', zscore=3)
nansd95 = core.partial_func(nansd, 'nansd95', ci=0.95)
nansd99 = core.partial_func(nansd, 'nansd99', ci=0.99)

se2 = core.partial_func(se, 'se2', zscore=2)
se3 = core.partial_func(se, 'se3', zscore=3)
se95 = core.partial_func(se, 'se95', ci=0.95)
se99 = core.partial_func(se, 'se99', ci=0.99)
nanse2 = core.partial_func(nanse, 'nanse2', zscore=2)
nanse3 = core.partial_func(nanse, 'nanse3', zscore=3)
nanse95 = core.partial_func(nanse, 'nanse95', ci=0.95)
nanse99 = core.partial_func(nanse, 'nanse99', ci=0.99)

mad2 = core.partial_func(mad, 'mad2', zscore=2)
mad3 = core.partial_func(mad, 'mad3', zscore=3)
mad95 = core.partial_func(mad, 'mad95', ci=0.95)
mad99 = core.partial_func(mad, 'mad99', ci=0.99)
nanmad2 = core.partial_func(nanmad, 'nanmad2', zscore=2)
nanmad3 = core.partial_func(nanmad, 'nanmad3', zscore=3)
nanmad95 = core.partial_func(nanmad, 'nanmad95', ci=0.95)
nanmad99 = core.partial_func(nanmad, 'nanmad99', ci=0.99)

######################
### Non-arrayfuncs ###
######################

def rstack(*arrays, sort_keys=False):
    """
    Stack the rows of multiple arrays.

    If scalar value(s) are given then these values will be appended to each column in the returned array.

    Parameters
    ----------
    arrays
        Arrays to be joined together.

    Returns
    -------
    IsopyArray
        Array containing all the row data and all the columns keys found in *arrays*.

    Examples
    --------
    >>> a = isopy.array(ru = 1, pd = 3, cd = 5)
    >>> b = isopy.array(rh = [2, 22], pd = [3, 33], ag = [4, 44])
    >>> isopy.rstack(a, b, 100)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)    Rh (f8)    Ag (f8)
    -------  ---------  ---------  ---------  ---------  ---------
    0          1.00000    3.00000    5.00000        nan        nan
    1              nan    3.00000        nan    2.00000    4.00000
    2              nan   33.00000        nan   22.00000   44.00000
    3        100.00000  100.00000  100.00000  100.00000  100.00000
    IsopyNdarray(2, flavour='element', default_value=nan)
    >>> isopy.rstack(a.default(0), b.default([0.1, 0.2]), [100, 200], sort_keys = True)
    (row)      Ru (f8)    Rh (f8)    Pd (f8)    Ag (f8)    Cd (f8)
    -------  ---------  ---------  ---------  ---------  ---------
    0          1.00000    0.00000    3.00000    0.00000    5.00000
    1          0.10000    2.00000    3.00000    4.00000    0.10000
    2          0.20000   22.00000   33.00000   44.00000    0.20000
    3        100.00000  100.00000  100.00000  100.00000  100.00000
    4        200.00000  200.00000  200.00000  200.00000  200.00000
    IsopyNdarray(4, flavour='element', default_value=nan)

    See Also
    --------
    rstack, cstack
    """
    arrays = [core.asanyarray(a) for a in arrays]

    keys = core.askeylist(*(a.keys for a in arrays if isinstance(a, core.IsopyArray)),
                        ignore_duplicates=True, sort=sort_keys)
    #arrays = [a.reshape(1) if a.ndim == 0 else a for a in arrays]

    for i, a in enumerate(arrays):
        if not isinstance(a, core.IsopyArray):
            arrays[i] = core.full(a.size, a, keys)

    result = [np.concatenate([a.get(key).reshape(1) if a.ndim == 0 else a.get(key) for a in arrays]) for key in keys]
    dtype = [(key, result[i].dtype) for i, key in enumerate(keys.strlist())]
    return keys._view_array_(np.fromiter(zip(*result), dtype=dtype))

def cstack(*arrays, sort_keys=False):
    """
    Stack the columns of multiple arrays.

    An error will thrown if a column occurs in more than one array.

    Normal numpy broadcasting rules apply so when concatenating columns the shape of the arrays must be compatible.
    When array with a size of 1 is concatenated to an array of a different size it will be copied into every row
    of the new shape.

    Parameters
    ----------
    arrays
        Arrays to be joined together.

    Returns
    -------
    IsopyArray
        Array containing all the row data and all the columns keys found in *arrays*.

    Examples
    --------
    >>> a = isopy.array(ru = 1, pd = 3, cd = 5)
    >>> b = isopy.array(rh = [2, 22],  ag = [4, 44])
    >>> isopy.cstack(a, b)
    (row)      Ru (f8)    Pd (f8)    Cd (f8)    Rh (f8)    Ag (f8)
    -------  ---------  ---------  ---------  ---------  ---------
    0          1.00000    3.00000    5.00000    2.00000    4.00000
    1          1.00000    3.00000    5.00000   22.00000   44.00000
    IsopyNdarray(2, flavour='element', default_value=nan)
    >>> isopy.cstack(a, b, sort_keys=True)
    (row)      Ru (f8)    Rh (f8)    Pd (f8)    Ag (f8)    Cd (f8)
    -------  ---------  ---------  ---------  ---------  ---------
    0          1.00000    2.00000    3.00000    4.00000    5.00000
    1          1.00000   22.00000    3.00000   44.00000    5.00000
    IsopyNdarray(2, flavour='element', default_value=nan)

    See Also
    --------
    rstack, cstack
    """
    arrays = [core.asarray(a) for a in arrays]
    size = {a.size for a in arrays}
    ndim = {a.ndim for a in arrays}

    if not (len(size) == 1 or (1 in size and len(size) == 2)):
        raise ValueError('Arrays have incompatible sizes and cannot be stacked')
    if len(ndim) != 1:
        arrays = [array.reshape(1) if array.ndim == 0 else array for array in arrays]

    keys = core.askeylist(*(a.keys for a in arrays), allow_duplicates=False, sort=sort_keys)

    result = {}
    size = _max(size) * _max(ndim) or None
    for key in keys:
        for a in arrays:
            if key in a.keys():
                result[key] = np.full(size, a.get(key))

    return core.array(result)

def concatenate(*arrays, axis=0):
    """
    Join arrays together along specified axis.

    If *arrays* contains at least one isopy array the input is forwarded to *rstack*/*cstack*. Otherwise the
    input is forwarded to *np.concatenate*.

    Parameters
    ----------
    arrays
        Arrays to be joined together.
    axis
        If 0 *rstack* is used. If 1 *cstack* is used.

    Returns
    -------
    IsopyArray
        Array containing all the row data and all the columns keys found in *arrays*.

    See Also
    --------
    rstack, cstack
    """
    arrays = [core.asanyarray(a) for a in arrays]

    if True not in [isinstance(a, core.IsopyArray) for a in arrays]:
        return np.concatenate([a for a in arrays if a is not None], axis=axis)

    if axis == 0 or axis is None:
        return rstack(*arrays)
    elif axis == 1:
        return cstack(*arrays)
    else:
        raise np.AxisError(axis, 1, 'concatenate only accepts axis values of 0 or 1')

def add(a1, a2, *, keys = None, **kwargs):
    """
    Compute ``a1 + a2``.

    Same as calling ``arrayfunc(np.add, *args, keys=keys, **kwargs)``. See :func:`arrayfunc` for a more
    detailed description on the workings of this function.

    Parameters
    ----------
    a1, a2 : array_like, numpy_array_like
        Both inputs must have the same shape or at least one of the inputs must have a size of 1.
    keys
        If given then the this will be the keys of the returned array/dictionary.
    kwargs
        Can either be a valid kwargs for the numpy function or a key filter. If key filters are given
        then only the columns with matching keys will be used to compute the result.

    Examples
    --------
    >>> a1 = isopy.array(ru = 1, pd = 3, ag = 4)
    >>> a2 = isopy.array(ru = 1, rh=2, pd=3)
    >>> isopy.add(a1, a2) # same as: a1 + a2
    (row)      Ru (f8)    Rh (f8)    Pd (f8)    Ag (f8)
    -------  ---------  ---------  ---------  ---------
    None       2.00000        nan    6.00000        nan
    IsopyNdarray(-1, flavour='element', default_value=nan)

    >>> d2 = a2.to_refval()
    >>> isopy.add(a1, d2) # same as: a1 + d2
    (row)      Ru (f8)    Pd (f8)    Ag (f8)
    -------  ---------  ---------  ---------
    None       2.00000    6.00000        nan
    IsopyNdarray(-1, flavour='element', default_value=nan)

    >>> isopy.add(1, 2, keys='ru pd cd')
    (row)      Ru (i8)    Pd (i8)    Cd (i8)
    -------  ---------  ---------  ---------
    None             3          3          3
    IsopyNdarray(-1, flavour='element', default_value=nan)
    """
    return arrayfunc(np.add, a1, a2, keys=keys, **kwargs)

def subtract(a1, a2, *, keys = None, **kwargs):
    """
    Compute ``a1 - a2``.

    Same as calling ``arrayfunc(np.add, *args, keys=keys, **kwargs)``. See :func:`arrayfunc` for a more
    detailed description on the workings of this function.

    Parameters
    ----------
    a1, a2 : array_like, numpy_array_like
        Both inputs must have the same shape or at least one of the inputs must have a size of 1.
    keys
        If given then the this will be the keys or the returned array.
    kwargs
        Can either be a valid kwargs for the numpy function or a key filter. If key filters are given
        then only the columns with matching keys will be used to compute the result.

    Examples
    --------
    >>> a1 = isopy.array(ru = 1, pd = 3, ag = 4)
    >>> a2 = isopy.array(ru = 1, rh=2, pd=3)
    >>> isopy.subtract(a1, a2) # same as: a1 - a2
    (row)      Ru (f8)    Rh (f8)    Pd (f8)    Ag (f8)
    -------  ---------  ---------  ---------  ---------
    None       0.00000        nan    0.00000        nan
    IsopyNdarray(-1, flavour='element', default_value=nan)

    >>> d2 = a2.to_refval()
    >>> isopy.subtract(a1, d2) # same as: a1 - d2
    (row)      Ru (f8)    Pd (f8)    Ag (f8)
    -------  ---------  ---------  ---------
    None       0.00000    0.00000        nan
    IsopyNdarray(-1, flavour='element', default_value=nan)

    >>> isopy.subtract(1, 2, keys='ru pd cd')
    (row)      Ru (i8)    Pd (i8)    Cd (i8)
    -------  ---------  ---------  ---------
    None            -1         -1         -1
    IsopyNdarray(-1, flavour='element', default_value=nan)
    """
    return arrayfunc(np.subtract, a1, a2, keys=keys, **kwargs)

def multiply(a1, a2, *, keys = None, **kwargs):
    """
    Compute ``a1 * a2``.

    Same as calling ``arrayfunc(np.add, *args, keys=keys, **kwargs)``. See :func:`arrayfunc` for a more
    detailed description on the workings of this function.

    Parameters
    ----------
    a1, a2 : array_like, numpy_array_like
        Both inputs must have the same shape or at least one of the inputs must have a size of 1.
    keys
        If given then the this will be the keys or the returned array.
    kwargs
        Can either be a valid kwargs for the numpy function or a key filter. If key filters are given
        then only the columns with matching keys will be used to compute the result.

    Examples
    --------
    >>> a1 = isopy.array(ru = 1, pd = 3, ag = 4)
    >>> a2 = isopy.array(ru = 1, rh=2, pd=3)
    >>> isopy.multiply(a1, a2) # same as: a1 * a2
    (row)      Ru (f8)    Rh (f8)    Pd (f8)    Ag (f8)
    -------  ---------  ---------  ---------  ---------
    None       1.00000        nan    9.00000        nan
    IsopyNdarray(-1, flavour='element', default_value=nan)

    >>> d2 = a2.to_refval()
    >>> isopy.multiply(a1, d2) # same as: a1 * d2
    (row)      Ru (f8)    Pd (f8)    Ag (f8)
    -------  ---------  ---------  ---------
    None       1.00000    9.00000        nan
    IsopyNdarray(-1, flavour='element', default_value=nan)

    >>> isopy.multiply(1, 2, keys='ru pd cd')
    (row)      Ru (i8)    Pd (i8)    Cd (i8)
    -------  ---------  ---------  ---------
    None             2          2          2
    IsopyNdarray(-1, flavour='element', default_value=nan)
    """
    return arrayfunc(np.multiply, a1, a2, keys=keys, **kwargs)

def divide(a1, a2, *, keys = None, **kwargs):
    """
    Compute ``a1 / a2``.

    Same as calling ``arrayfunc(np.add, *args, keys=keys, **kwargs)``. See :func:`arrayfunc` for a more
    detailed description on the workings of this function.

    Parameters
    ----------
    a1, a2 : array_like, numpy_array_like
        Both inputs must have the same shape or at least one of the inputs must have a size of 1.
    keys
        If given then the this will be the keys or the returned array.
    kwargs
        Can either be a valid kwargs for the numpy function or a key filter. If key filters are given
        then only the columns with matching keys will be used to compute the result.

    Examples
    --------
    >>> a1 = isopy.array(ru = 1, pd = 3, ag = 4)
    >>> a2 = isopy.array(ru = 1, rh=2, pd=3)
    >>> isopy.divide(a1, a2) # same as: a1 / a2
    (row)      Ru (f8)    Rh (f8)    Pd (f8)    Ag (f8)
    -------  ---------  ---------  ---------  ---------
    None       1.00000        nan    1.00000        nan
    IsopyNdarray(-1, flavour='element', default_value=nan)

    >>> d2 = a2.to_refval()
    >>> isopy.divide(a1, d2) # same as: a1 / d2
    (row)      Ru (f8)    Pd (f8)    Ag (f8)
    -------  ---------  ---------  ---------
    None       1.00000    1.00000        nan
    IsopyNdarray(-1, flavour='element', default_value=nan)


    >>> isopy.divide(1, 2, keys='ru pd cd')
    (row)      Ru (f8)    Pd (f8)    Cd (f8)
    -------  ---------  ---------  ---------
    None       0.50000    0.50000    0.50000
    IsopyNdarray(-1, flavour='element', default_value=nan)
    """
    return arrayfunc(np.divide, a1, a2, keys=keys, **kwargs)

def power(a1, a2, *, keys = None, **kwargs):
    """
    Compute ``a1 ** a2``.

    Same as calling ``arrayfunc(np.add, *args, keys=keys, **kwargs)``. See :func:`arrayfunc` for a more
    detailed description on the workings of this function.

    Parameters
    ----------
    a1, a2 : array_like, numpy_array_like
        Both inputs must have the same shape or at least one of the inputs must have a size of 1.
    keys
        If given then the this will be the keys or the returned array.
    kwargs
        Can either be a valid kwargs for the numpy function or a key filter. If key filters are given
        then only the columns with matching keys will be used to compute the result.

    Examples
    --------
    >>> a1 = isopy.array(ru = 1, pd = 3, ag = 4)
    >>> a2 = isopy.array(ru = 1, rh=2, pd=3)
    >>> isopy.power(a1, a2) # same as: a1 ** a2
    (row)      Ru (f8)    Rh (f8)    Pd (f8)    Ag (f8)
    -------  ---------  ---------  ---------  ---------
    None       1.00000        nan   27.00000        nan
    IsopyNdarray(-1, flavour='element', default_value=nan)

    >>> d2 = a2.to_refval()
    >>> isopy.power(a1, d2) # same as: a1 ** d2
    (row)      Ru (f8)    Pd (f8)    Ag (f8)
    -------  ---------  ---------  ---------
    None       1.00000   27.00000        nan
    IsopyNdarray(-1, flavour='element', default_value=nan)



    >>> isopy.power(1, 2, keys='ru pd cd')
    (row)      Ru (i8)    Pd (i8)    Cd (i8)
    -------  ---------  ---------  ---------
    None             1          1          1
    IsopyNdarray(-1, flavour='element', default_value=nan)
    """
    return arrayfunc(np.power, a1, a2, keys=keys, **kwargs)

@core.append_preset_docstring
@core.add_preset('sd', cval=np.mean, pmval=sd)
@core.add_preset('sd2', cval=np.mean, pmval=sd2)
@core.add_preset('sd3', cval=np.mean, pmval=sd3)
@core.add_preset('se', cval=np.mean, pmval=se)
@core.add_preset('se2', cval=np.mean, pmval=se2)
@core.add_preset('se3', cval=np.mean, pmval=se3)
@core.add_preset('mad', cval=np.median, pmval=mad)
@core.add_preset('mad2', cval=np.median, pmval=mad2)
@core.add_preset('mad3', cval=np.median, pmval=mad3)
def is_outlier(data, cval = np.median, pmval=mad3, axis = None):
    """
    Mark all the outliers in the data as ``True`` and the rest as ``False``.

    Parameters
    ----------
    data : isopy_array_like
        The outliers will be calculated for each column in *data*.
    cval : scalar, Callable
        Either the center value or a function that returns the center
        value when called with *data*.
    pmval : scalar, Callable
        Either the uncertainty value or a function that returns the
        uncertainty when called with *data*.
    axis : {0, 1}, Optional
        If not given then an array with each individual outlier marked is returned. Otherwise
        ``np.any(outliers, axis)`` is returned.
    """
    if callable(cval):
        cval = cval(data)
    if callable(pmval):
        pmval = pmval(data)
    pmval = np.abs(pmval)

    outliers = (data > (cval + pmval)) + (data < (cval - pmval))

    if axis is not None:
        outliers = np.any(outliers, axis=axis)

    return outliers

@core.append_preset_docstring
@core.add_preset('sd', cval=np.mean, pmval=sd)
@core.add_preset('sd2', cval=np.mean, pmval=sd2)
@core.add_preset('sd3', cval=np.mean, pmval=sd3)
@core.add_preset('se', cval=np.mean, pmval=se)
@core.add_preset('se2', cval=np.mean, pmval=se2)
@core.add_preset('se3', cval=np.mean, pmval=se3)
@core.add_preset('mad', cval=np.median, pmval=mad)
@core.add_preset('mad2', cval=np.median, pmval=mad2)
@core.add_preset('mad3', cval=np.median, pmval=mad3)
def not_outlier(data, cval = np.median, pmval=mad3, axis = None):
    """
    Mark all the outliers in the data as ``False`` and the rest as ``True``.

    The inverted value of :func:`is_outlier`.

    Parameters
    ----------
    data : isopy_array_like
        The outliers will be calculated for each column in *data*.
    cval : scalar, Callable
        Either the center value or a function that returns the center
        value when called with *data*.
    pmval : scalar, Callable
        Either the uncertainty value or a function that returns the
        uncertainty when called with *data*.
    axis : {0, 1}, Optional
        If not given then an array with each individual outlier marked is returned. Otherwise
        ``np.any(outliers, axis)`` is returned.
    """
    outliers = is_outlier(data, cval, pmval, axis)
    return np.invert(outliers)

@core.append_preset_docstring
@core.add_preset('sd', cval=np.mean, pmval=sd)
@core.add_preset('sd2', cval=np.mean, pmval=sd2)
@core.add_preset('sd3', cval=np.mean, pmval=sd3)
@core.add_preset('se', cval=np.mean, pmval=se)
@core.add_preset('se2', cval=np.mean, pmval=se2)
@core.add_preset('se3', cval=np.mean, pmval=se3)
@core.add_preset('mad', cval=np.median, pmval=mad)
@core.add_preset('mad2', cval=np.median, pmval=mad2)
@core.add_preset('mad3', cval=np.median, pmval=mad3)
def upper_limit(data, cval=np.median, pmval=mad3):
    """
    Calculate the upper limit of the uncertainty on *data* as ``cval + pmval``.

    Parameters
    ----------
    data
        The data on which the limit will be calculated
    cval
        The centre value. Can either be a scalar value or a function that returns the centre value.
    pmval
        The uncertainty around *cval*. Can either be a scalar value or a function that returns the uncertainty.
    """
    if callable(cval):
        cval = cval(data)
    if callable(pmval):
        pmval = pmval(data)
    pmval = np.abs(pmval)

    return cval + pmval

@core.append_preset_docstring
@core.add_preset('sd', cval=np.mean, pmval=sd)
@core.add_preset('sd2', cval=np.mean, pmval=sd2)
@core.add_preset('sd3', cval=np.mean, pmval=sd3)
@core.add_preset('se', cval=np.mean, pmval=se)
@core.add_preset('se2', cval=np.mean, pmval=se2)
@core.add_preset('se3', cval=np.mean, pmval=se3)
@core.add_preset('mad', cval=np.median, pmval=mad)
@core.add_preset('mad2', cval=np.median, pmval=mad2)
@core.add_preset('mad3', cval=np.median, pmval=mad3)
def lower_limit(data, cval=np.median, pmval=mad3):
    """
    Calculate the lower limit of the uncertainty on *data* as ``cval + pmval``.

    Parameters
    ----------
    data
        The data on which the limit will be calculated
    cval
        The centre value. Can either be a scalar value or a function that returns the centre value.
    pmval
        The uncertainty around *cval*. Can either be a scalar value or a function that returns the uncertainty.
    """
    if callable(cval):
        cval = cval(data)
    if callable(pmval):
        pmval = pmval(data)
    pmval = np.abs(pmval)

    return cval - pmval

@core.append_preset_docstring
@core.add_preset('abs', abs = True)
def keymax(a, evalfunc = np.nanmedian, abs=False):
    """
    Return the name of the column where the largest value returned by *evalfunc* is found.

    If *abs* is ``True`` the absolute value is used.

    This function requires that *a* is an isopy array.

    Examples
    --------
    >>> array = isopy.tb.make_ms_array('pd')
    >>> isopy.keymax(array)
    IsotopeKeyString('106Pd')
    """
    array = core.asarray(a)
    if abs:
        key_index = np.nanargmax([evalfunc(np.abs(v)) for v in array.values()])
    else:
        key_index = np.nanargmax([evalfunc(v) for v in array.values()])
    return array.keys[key_index]

@core.append_preset_docstring
@core.add_preset('abs', abs = True)
def keymin(a, evalfunc= np.nanmedian, abs = False):
    """
    Return the name of the column where the smallest value returned by *evalfunc* is found.

    If *abs* is ``True`` the absolute value is used.

    This function requires that *a* is an isopy array.

    Examples
    --------
    >>> array = isopy.tb.make_ms_array('pd')
    >>> isopy.keymin(array)
    IsotopeKeyString('102Pd')
    """
    array = core.asarray(a)
    if abs:
        key_index = np.nanargmin([evalfunc(np.abs(v)) for v in array.values()])
    else:
        key_index = np.nanargmin([evalfunc(v) for v in array.values()])
    return array.keys[key_index]

################################
### Approved numpy functions ###
################################

np_elementwise = [np.sin, np.cos, np.tan, np.arcsin, np.arccos, np.arctan, np.degrees, np.isnan,
                  np.radians, np.deg2rad, np.rad2deg, np.sinh, np.cosh, np.tanh, np.arcsinh,
                  np.arccosh, np.arctanh,
                  np.rint, np.floor, np.ceil, np.trunc, np.exp, np.expm1, np.exp2, np.log,
                  np.log10, np.log2,
                  np.log1p, np.reciprocal, np.positive, np.negative, np.sqrt, np.cbrt, np.square,
                  np.fabs, np.sign,
                  np.absolute, np.abs]
np_cumulative = [np.cumprod, np.cumsum, np.nancumprod, np.nancumsum]
np_reducing = [np.prod, np.sum, np.nanprod, np.nansum, np.cumprod, np.cumsum, np.nancumprod, np.nancumsum,
               np.amin, np.amax, np.nanmin, np.nanmax, np.ptp, np.median, np.average, np.mean,
               np.std, np.var, np.nanmedian, np.nanmean, np.nanstd, np.nanvar, np.nanmax, np.nanmin,
               np.all, np.any]
np_special = [np.copyto, np.average]
np_dual = [np.add, np.subtract, np.divide, np.multiply, np.power]


for functions in [np_elementwise, np_cumulative, np_reducing, np_special, np_dual]:
    for func in functions:
        new_func = arrayfunc_wrapper(func)
        core.APPROVED_FUNCTIONS.append(func)
        if func.__name__ not in __all__:
            __all__.append(func.__name__)
            globals()[func.__name__] = new_func
        if func.__name__ == 'amin':
            __all__.append('min')
            globals()['min'] = new_func
        if func.__name__ == 'amax':
            __all__.append('max')
            globals()['max'] = new_func
        if func.__name__ == 'absolute':
            __all__.append('abs')
            globals()['abs'] = new_func

def approved_numpy_functions(format ='name', delimiter =', '):
    """
    Return a string containing the names of all the numpy functions supported by isopy arrays.

    With *format* you can specify the format of the string for each function. Avaliable keywords
    are ``{name}`` and ``{link}`` for the name and a link the numpy documentation web page for a
    function. Avaliable presets are ``"name"``, ``"link"``, ``"rst"`` and ``"markdown"`` for just the name,
    just the link, reStructured text hyper referenced link or a markdown hyper referenced link.

    You can specify the delimiter used to seperate items in the list using the *delimiter* argument.
    If *delimiter* is ``None`` a python list is returned.
    """
    if format == 'name': format = '{name}'
    if format == 'link': format = '{link}'
    if format == 'rst': format = '`{name} <{link}>`_'
    if format == 'markdown': format = '[{name}]({link})'

    strings = []
    for functions in [np_elementwise, np_cumulative, np_reducing, np_special, np_dual]:
        for func in functions:
            name = func.__name__
            link = f'https://numpy.org/doc/stable/reference/generated/numpy.{name}.html'
            strings.append(format.format(name = name, link=link))
    if delimiter is None:
        return strings
    else:
        return delimiter.join(strings)